/******************************************************************************
*******************************************************************************
**                                                                           **
**      C Swing - A graphical file and directory maintenance utility         **
**                                                                           **
*******************************************************************************
**                                                                           **
**      Original C coding by: S. J. Brown                                    **
**                                                                           **
**      Modifications by many others; see the documentation for a list       **
**                                                                           **
**      Copyright 1994, Harry Flowers and Foteos Macrides                    **
**                                                                           **
**      Terms of Copyright:  This software may be freely distributed         **
**          and used by anyone, providing the distributor provides it        **
**          free of charge or at a price which only covers distribution      **
**          costs.  It may not be sold.  Note that versions prior to         **
**          V3.7.4 were public domain, and are not subject to these terms.   **
**                                                                           **
**      Disclaimer: No warranties as to the use or reliability of this       **
**          software are given by the authors or distributors.  Use at       **
**          your own risk.                                                   **
**                                                                           **
*******************************************************************************
******************************************************************************/
 
/*
**  swing_14 - a real rename function using RMS
*/
#include <stdio.h>
#include <iodef.h>
#include <rms.h>
#include <ssdef.h>
#include <atrdef.h>
#include <fibdef.h>
#include "more_descrip.h"
#include "swing.h"
#include "swing_refs.h"

static short fpro;
static int uic;
static struct fibdef owner_fib, prot_fib, blank_fib;

char revdate[ATR$S_REVDATE], 
     ascdates[ATR$S_ASCDATES];

ACB acb_pro[] = {
               /*
               **  protection
               */
               ATR$S_FPRO,	ATR$C_FPRO,	&fpro,

               /*
               **  original revision date - need to rewrite this at the
               **  end 
               */
               ATR$S_REVDATE,   ATR$C_REVDATE,   revdate,

               /*
               **  revision count contained in first two bytes of ascdates
               */
               2,  		ATR$C_ASCDATES,  ascdates,

               /*
               **  terminate the list
               */
               0,               0,               0};

ACB acb_own[] = {
               /*
               **  ownership
               */
               ATR$S_UIC,	ATR$C_UIC,	&uic,

               /*
               **  original revision date - need to rewrite this at the
               **  end 
               */
               ATR$S_REVDATE,   ATR$C_REVDATE,   revdate,

               /*
               **  revision count contained in first two bytes of ascdates
               */
               2,  		ATR$C_ASCDATES,  ascdates,

               /*
               **  terminate the list
               */
               0,               0,               0};

lib_rename(from,to,name)
char *from,*to,*name;
{
    char esa1[SPEC_SIZE+1], esa2[SPEC_SIZE+2];
    char rsa[SPEC_SIZE+1], dss[SPEC_SIZE+1];

    int i,len,status;
    short prot, dummy;

    struct FAB fab1, fab2, fab;
    struct NAM nam1, nam2;

    fab1 = cc$rms_fab;
    fab2 = cc$rms_fab;
    nam1 = cc$rms_nam;
    nam2 = cc$rms_nam;

    /*
    **  find the default filespec
    */
    strcpy(dss,from);
    len = strlen(dss);
    for (i = 0; i < len; i++)
        if (dss[i] == ';')
        {
            dss[i] = 0;
            break;
        }
    /*
    **  fab1, nam1 used for old filespec
    */
    fab1.fab$l_fna = from;
    fab1.fab$b_fns = strlen(from);
    fab1.fab$l_nam = &nam1;
    nam1.nam$l_esa = esa1;
    nam1.nam$b_ess = SPEC_SIZE;

    /*
    **  fab2, nam2 used for new filespec
    */
    fab2.fab$l_fop = FAB$M_MXV;
    fab2.fab$l_dna = dss;
    fab2.fab$b_dns = strlen(dss);
    fab2.fab$l_fna = to;
    fab2.fab$b_fns = strlen(to);
    fab2.fab$l_nam = &nam2;
    nam2.nam$l_esa = esa2;
    nam2.nam$b_ess = SPEC_SIZE;
    nam2.nam$l_rsa = rsa;
    nam2.nam$b_rss = SPEC_SIZE;

    i = sys$rename(&fab1,0,0,&fab2);

    if (i != RMS$_SUC)
    {
        /*
        **  now change the protection to give W:D - if file has it return
        **  the rename error code - we can't do any more than we have
        */
	set_protection("SYS$DISK",nam1.nam$w_fid,0,&prot);
    
        /*
        **  let's try to rename it now
        */
        i = sys$rename(&fab1,0,0,&fab2);

        /*
        **  set protection back to what it was
        */
	set_protection("SYS$DISK",nam1.nam$w_fid,prot,&dummy);

        if (i != RMS$_SUC)
            return(i);
    }

    if (name)
        sprintf(name,"%.*s",nam2.nam$b_name+nam2.nam$b_type+nam2.nam$b_ver,
               nam2.nam$l_name);
    return(SS$_NORMAL);
}

/*
**  change file fid on device protection - return original protection
*/
set_protection(device,fid,protection,original)
char *device;
unsigned short fid[3];
unsigned short protection;
unsigned short *original;
{
    char dev[SPEC_SIZE+1];
    int i,l;
    int dir_channel;

    strcpy(dev,device);
    l = strlen(dev);
    for (i = 0; i < l; i++)
    {
	if (dev[i] == '[')
	{
	    dev[i] = 0;
	    break;
	}
    }

    dir_channel = 0;
    sys$assign(DSC1(dev),
               &dir_channel,
               0,
               0);

    change_the_protection(dir_channel,fid,protection,original);

    sys$dassgn(dir_channel);
}

change_the_protection(channel,fid,protection,original)
int channel;
unsigned short fid[3];
unsigned short protection;
unsigned short *original;
{
    int iosb[2],status;
    struct dsc$descriptor_s fib_desc;

    prot_fib = blank_fib;
        
    fib_desc.dsc$w_length  = FIB$C_LENGTH;
    fib_desc.dsc$a_pointer = &prot_fib;
    fib_desc.dsc$b_class   = DSC$K_CLASS_S;
    fib_desc.dsc$b_dtype   = DSC$K_DTYPE_T;
        
#ifdef __DECC
    prot_fib.fib$w_fid[0] = fid[0];
    prot_fib.fib$w_fid[1] = fid[1];
    prot_fib.fib$w_fid[2] = fid[2];
#else
    prot_fib.fib$r_fid_overlay.fib$w_fid[0] = fid[0];
    prot_fib.fib$r_fid_overlay.fib$w_fid[1] = fid[1];
    prot_fib.fib$r_fid_overlay.fib$w_fid[2] = fid[2];
#endif
    
    sys$qiow(0,
             channel,
             IO$_ACCESS,
             &iosb[0],
             0,
             0,
             &fib_desc,
             0,
             0,
             0,
             acb_pro,
             0);

    *original = fpro;

    /*
    **  no point in changing protection to what it already is
    */
    if (protection == *original)
	return(SS$_NORMAL);

    prot_fib = blank_fib;

#ifdef __DECC
    prot_fib.fib$w_fid[0] = fid[0];
    prot_fib.fib$w_fid[1] = fid[1];
    prot_fib.fib$w_fid[2] = fid[2];
    
    prot_fib.fib$l_acctl = FIB$M_WRITE;
#else
    prot_fib.fib$r_fid_overlay.fib$w_fid[0] = fid[0];
    prot_fib.fib$r_fid_overlay.fib$w_fid[1] = fid[1];
    prot_fib.fib$r_fid_overlay.fib$w_fid[2] = fid[2];
    
    prot_fib.fib$r_acctl_overlay.fib$l_acctl = FIB$M_WRITE;
#endif

    fpro = protection;

    status = sys$qiow(0,
             channel,
             IO$_MODIFY,
             &iosb[0],
             0,
             0,
             &fib_desc,
             0,
             0,
             0,
             acb_pro,
             0);
    if (status == SS$_NORMAL)
	status = iosb[0];

    return(status);
}

change_the_ownership(channel,fid,owner)
int channel;
unsigned short fid[3];
unsigned int owner;
{
    int iosb[2],status;
    struct dsc$descriptor_s fib_desc;

    owner_fib = blank_fib;
        
    fib_desc.dsc$w_length  = FIB$C_LENGTH;
    fib_desc.dsc$a_pointer = &owner_fib;
    fib_desc.dsc$b_class   = DSC$K_CLASS_S;
    fib_desc.dsc$b_dtype   = DSC$K_DTYPE_T;
        
#ifdef __DECC
    owner_fib.fib$w_fid[0] = fid[0];
    owner_fib.fib$w_fid[1] = fid[1];
    owner_fib.fib$w_fid[2] = fid[2];
#else
    owner_fib.fib$r_fid_overlay.fib$w_fid[0] = fid[0];
    owner_fib.fib$r_fid_overlay.fib$w_fid[1] = fid[1];
    owner_fib.fib$r_fid_overlay.fib$w_fid[2] = fid[2];
#endif
    
    sys$qiow(0,
             channel,
             IO$_ACCESS,
             &iosb[0],
             0,
             0,
             &fib_desc,
             0,
             0,
             0,
             acb_own,
             0);

    if (uic == owner)
	return(SS$_NORMAL);

    owner_fib = blank_fib;
        
    fib_desc.dsc$w_length  = FIB$C_LENGTH;
    fib_desc.dsc$a_pointer = &owner_fib;
    fib_desc.dsc$b_class   = DSC$K_CLASS_S;
    fib_desc.dsc$b_dtype   = DSC$K_DTYPE_T;
        
#ifdef __DECC
    owner_fib.fib$w_fid[0] = fid[0];
    owner_fib.fib$w_fid[1] = fid[1];
    owner_fib.fib$w_fid[2] = fid[2];
    
    owner_fib.fib$l_acctl = FIB$M_WRITE;
#else
    owner_fib.fib$r_fid_overlay.fib$w_fid[0] = fid[0];
    owner_fib.fib$r_fid_overlay.fib$w_fid[1] = fid[1];
    owner_fib.fib$r_fid_overlay.fib$w_fid[2] = fid[2];
    
    owner_fib.fib$r_acctl_overlay.fib$l_acctl = FIB$M_WRITE;
#endif

    uic = owner;

    status = sys$qiow(0,
             channel,
             IO$_MODIFY,
             &iosb[0],
             0,
             0,
             &fib_desc,
             0,
             0,
             0,
             acb_own,
             0);
    if (status == SS$_NORMAL)
	status = iosb[0];

    return(status);
}

