/******************************************************************************
*******************************************************************************
**                                                                           **
**      C Swing - A graphical file and directory maintenance utility         **
**                                                                           **
*******************************************************************************
**                                                                           **
**      Original C coding by: S. J. Brown                                    **
**                                                                           **
**      Modifications by many others; see the documentation for a list       **
**                                                                           **
**      Copyright 1994, Harry Flowers and Foteos Macrides                    **
**                                                                           **
**      Terms of Copyright:  This software may be freely distributed         **
**          and used by anyone, providing the distributor provides it        **
**          free of charge or at a price which only covers distribution      **
**          costs.  It may not be sold.  Note that versions prior to         **
**          V3.7.4 were public domain, and are not subject to these terms.   **
**                                                                           **
**      Disclaimer: No warranties as to the use or reliability of this       **
**          software are given by the authors or distributors.  Use at       **
**          your own risk.                                                   **
**                                                                           **
*******************************************************************************
******************************************************************************/
 
/* 
**  swing_30 file manager cursor bashing routines
*/
#include <stdio.h>
#include <ssdef.h>
#include <smgdef.h>
#include "more_descrip.h"
#include "swing.h"
#include "swing_refs.h"

clear_cursor()
{
    int status;

    if (cursor_line < 0)
        return(TRUE);

    status = smg$put_chars(&displayf,DSC1("  "),&cursor_line,&1);
    if (status != SS$_NORMAL)
	panic(status,"put_chars 1","clear curor");
}

set_cursor(line)
int line;
{
    int status;

    cursor_line = line;

    status = smg$put_chars(&displayf,DSC1("->"),&cursor_line,
                         &1,0,&CURSOR_RENDITION);
    if (status != SS$_NORMAL)
	panic(status,"put_chars 1","set cursor");
}

put_top_files(count,top_line,display_type)
int count;
int top_line;
int display_type;
{
    FILER *temp;
    int i,status,display_lines;

    if (display_type == FILE_MANAGER)
        display_lines = DISPLAYF_FILES;
    else
        display_lines = DISPLAYF_ROWS;

    temp = current_file_ptr(top_line+count-1);

    for (i = 0; i < count; i++)
    {
        status = smg$scroll_display_area(&displayf,
                                         &1,&4,
                                         &display_lines,
                                         &(SCROLL_WIDTH),
                                         &SMG$M_DOWN);
        if (status != SS$_NORMAL)
            panic(status,"scroll display area","put file top");

        put_scroll_line(1,temp,display_type);
	temp = temp->back_ptr;
    }
}

put_bottom_files(count,top_line,display_type)
int count;
int top_line;
int display_type;
{
    FILER *temp;
    int i,status,display_lines;

    if (display_type == FILE_MANAGER)
        display_lines = DISPLAYF_FILES;
    else
        display_lines = DISPLAYF_ROWS;

    temp = current_file_ptr(top_line + display_lines - count);

    for (i = 0; i < count; i++)
    {
        status = smg$scroll_display_area(&displayf,
                                         &1,&4,
                                         &display_lines,
                                         &(SCROLL_WIDTH),
                                         &SMG$M_UP);
        if (status != SS$_NORMAL)
            panic(status,"scroll display area","put file bottom");

        put_scroll_line(display_lines,temp,display_type);
	temp = temp->forw_ptr;
    }
}

show_files(changed,display_type)
int changed;
int display_type;
{
    int i, d_line, length, status;
    int display_lines;

    if (display_type == FILE_MANAGER)
        display_lines = DISPLAYF_FILES;
    else
        display_lines = DISPLAYF_ROWS;

    if (curr_file_no >= total_files)
        curr_file_no = (total_files - 1);

    if (curr_file_no < 0)
        curr_file_no = 0;

    if (curr_top_file > (total_files - display_lines))
        curr_top_file = total_files - display_lines;

    if (curr_top_file < 0)
        curr_top_file = 0;

    smg$set_cursor_mode(&paste,&1);
    smg$begin_display_update(&displayf);
    clear_cursor();

    if (curr_file_no < curr_top_file)
    {
        i = 0;
        while(curr_file_no < (curr_top_file+(display_lines/2)) && curr_top_file)
        {
	    i++;
	    curr_top_file--;
        }
        if (i || changed)
        {
            if (i > MAX_FILES_SCROLL || changed)
		refresh_files(display_type);
	    else
    	        put_top_files(i,curr_top_file,display_type);
	}
    }
    else if (curr_file_no >= (curr_top_file+display_lines))
    {
	i = 0;
	while(curr_file_no >= (curr_top_file+(display_lines/2)) &&
	      curr_top_file < (total_files-display_lines))
        {
	    i++;
	    curr_top_file++;
	}
        if (i || changed)
        {
            if (i > MAX_FILES_SCROLL || changed)
		refresh_files(display_type);
	    else
    	        put_bottom_files(i,curr_top_file,display_type);
	}
    }
    else if (changed)
        refresh_files(display_type);

    set_cursor(curr_file_no-curr_top_file+1);

    i = display_lines + 1;

    smg$draw_line(&displayf,&1,&LEFTF_COL,&i,&LEFTF_COL);

    if (display_type == FILE_MANAGER)
    {
        d_line = DISPLAYF_FILES + 1;
        smg$draw_line(&displayf,&d_line,&1,&d_line,&DISPLAYF_COLUMNS);
    
        d_line++;
        smg$set_cursor_abs(&displayf,&d_line,&2);
        sprintf(junk,"current file no %-5d ",curr_file_no+1);
        sprintf(&junk[strlen(junk)],"total files %-5d ",total_files);
        sprintf(&junk[strlen(junk)],"files selected %-5d",files_selected);
        sprintf(&junk[strlen(junk)],"display format %d",fm_type);
        smg$put_line(&displayf,DSC1(junk),&1,0);
    }

    smg$end_display_update(&displayf);
    smg$set_cursor_mode(&paste,&0);
}

refresh_files(display_type)
int display_type;
{
    int i, status;
    int line, display_lines;

    if (display_type == FILE_MANAGER)
        display_lines = DISPLAYF_FILES;
    else
        display_lines = DISPLAYF_ROWS;

    curr_file_ptr = current_file_ptr(curr_top_file);
    smg$set_cursor_abs(&displayf,&1,&1);
    for (i = curr_top_file; i < curr_top_file + display_lines && 
                            i < total_files; i++)
    {
        line = i - curr_top_file + 1;

        put_scroll_line(line,curr_file_ptr,display_type);
        curr_file_ptr = curr_file_ptr->forw_ptr;
    }

    /*
    **  clear out rest of the display lines which had filenames on 'em
    */
    for (;i < curr_top_file + display_lines; i++)
    {
        line = i - curr_top_file + 1;

        sprintf(junk,"%-*.*s",SCROLL_WIDTH,SCROLL_WIDTH,"");
	status = smg$put_chars(&displayf,DSC1(junk),&line,&4);
        if (status != SS$_NORMAL)
            panic(status,"put chars","refresh files");
    }
}
